# -*- coding: utf-8 -*-

import sys
import utils

class Validator():

    def __init__(self, schema, OUTPUT_DIR=None):
        ''' Bind array of cleaned schema file lines to validator object. ''' 

        self.schema = utils.clean(schema)
        self.OUTPUT_DIR = OUTPUT_DIR
        self.INDENT_SIZE = self._find_first_indent()['indent_size']
        self.error = {'msg': None}

    def validate(self):
        ''' Return True if schema is valid, otherwise return False. '''

        if not len(self.schema): 
            self.error['msg'] = 'Error: empty schema file.'
            return False

        start_index = self._find_first_indent()['index']
        prev_line = self.schema[start_index]
        prev_indent = self.INDENT_SIZE

        schema_is_valid = True

        lines_to_validate = self.schema[start_index + 1:]
        for index, this_line in enumerate( lines_to_validate ):

            this_indent = utils.parse_indent(this_line)
            if not self._line_is_valid(this_line, prev_line, this_indent, prev_indent):

                self.error['msg'] = 'Error in schema file on line {}'.format((index + start_index + 1))
                schema_is_valid = False
                break

            prev_indent = this_indent
            prev_line = this_line


        if not self._top_dir_is_valid(self.OUTPUT_DIR):
            self.error['msg'] = 'Error regarding top-level directory rules. See superdir --help for usage.'
            schema_is_valid = False

        return schema_is_valid


    def _line_is_valid(self, this_line, prev_line, this_indent, prev_indent):
        '''
        Once the first indent size is determined, each subsequent
        indent must be:

            1) less than N by a multiple of N,  e.g. 8 -> 4 or 8 -> 0
            2) 0, or 
            3) preceded by a directory and greater than N by exactly N.
        '''

        difference = this_indent - prev_indent

        return  (difference == 0) or\
                (difference == self.INDENT_SIZE and utils.is_dir(prev_line)) or\
                (difference < 0 and utils.is_multiple_of_indent(this_indent, self.INDENT_SIZE)) 


    def _top_dir_is_valid(self, OUTPUT_DIR):
        ''' Return True only if output dir is supplied or schema file has exactly 1 top-level directory and no siblings '''

        indents = [ utils.parse_indent(line) for line in self.schema ]
        min_indent_count = indents.count( min(indents) ) 

        return OUTPUT_DIR is not None or min_indent_count == 1 

    def _find_first_indent(self):
        ''' Returns indent_value, start_index of first indent. '''

        indent_size, start_index = 0, 0 

        for index, line in enumerate(self.schema):
            if utils.parse_indent(line) > 0:
                indent_size = utils.parse_indent(line)
                start_index = index
                break

        return dict(indent_size=indent_size, index=start_index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import sys

import click

from tree import Tree
from validator import Validator
import callbacks as cbs
import utils


def main(schema=None, OUTPUT_DIR=None, CONFIG_PATH=None):
    ''' Validate schema file and output directory parameters, build a tree from schema, callback on each node.  '''

    validator = Validator(schema, OUTPUT_DIR=OUTPUT_DIR)
    if not validator.validate():
        click.echo(validator.error['msg'])
        sys.exit(1)

    directory_tree = Tree(
        INDENT_SIZE = validator.INDENT_SIZE,
        OUTPUT_DIR  = OUTPUT_DIR,
        base_path   = os.path.abspath(os.curdir)
    )
    directory_tree.load_data(schema)
    directory_tree.build_tree()

    callbacks = [
        cbs.create_file,
    ]

    if CONFIG_PATH:
        process_hooks = cbs.make_config_processor(CONFIG_PATH)
        callbacks.append(process_hooks)

    directory_tree.walk(callbacks=callbacks)


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.argument('schema_file', type=click.File('r'), required=True, default=sys.stdin)
@click.option('-c', '--config', nargs=1, type=str, help="Path to config file to read before superdir'ing your schema.")
@click.option('-o', '--outfile', nargs=1, type=str, help=("Directory name to contain your superdir'd files. If none is" 
                                                          "supplied, your schema file must have exactly one top-level directory"         
                                                          "and no sibling regular files. That top-level directory will be the"
                                                          "parent of your new file tree. If an output directory is supplied"
                                                          "then any validly-indented schema file is allowed."))
def cli(schema_file, outfile, config):

    schema = None

    if schema_file is None:

        # schema probably coming from a pipe
        if not sys.stdin.isatty():
            schema = list(sys.stdin) 

    else:

        # schema from a file
        schema = list(schema_file)

    main(schema=schema, OUTPUT_DIR=outfile, CONFIG_PATH=config)

if __name__ == '__main__':

    cli()



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import sys
import utils
import pdb

class Tree:

    def __init__(self, INDENT_SIZE=None, OUTPUT_DIR=None, base_path=None):

        self.INDENT_SIZE = INDENT_SIZE
        self.OUTPUT_DIR = OUTPUT_DIR
        self.base_path = base_path
        self.data = None
        self.root = None

    def build_tree(self):
        ''' Build tree from indentation. '''

        virtual_root = self._make_new_node(
            parent      = None,
            children    = [],
            value       = None,
            path        = self.base_path
        )

        root = self._make_new_node(
            parent      = virtual_root,
            children    = [],
            value       = self.OUTPUT_DIR, 
            # issue here: if no output dir? Then what ?
            path        = ( os.path.join(self.base_path, self.OUTPUT_DIR) 
                            if self.OUTPUT_DIR 
                            else self.base_path )
        )

        virtual_root['children'].append(root)

        parent_node = virtual_root
        prev_indent = -1
        for line in self.data:

            cur_indent = utils.get_indent_count(line, self.INDENT_SIZE)
            distance = cur_indent - prev_indent
            filename = (utils.get_dirname(line)
                       if utils.is_dir(line)
                       else utils.get_filename(line))

            parent_node = self._find_new_parent(parent_node, distance)

            child = self._make_new_node(
                parent   = parent_node, 
                children = [] if utils.is_dir(line) else None,
                value    = filename,
                path     = os.path.join(parent_node['path'], filename)
            ) 

            parent_node['children'].append(child)
            prev_indent = cur_indent

        self.root = virtual_root

    def _find_new_parent(self, parent_node, distance):

        new_parent = parent_node

        if distance > 0:
            new_parent = parent_node['children'][-1]

        elif distance < 0:
            new_parent = self._find_ancestor(parent_node, distance)

        return new_parent

    def walk(self, callbacks):
        ''' Walk tree and call callback on each node. '''

        def _walk(tree):

            children = tree['children']

            for child in children:

                if callbacks:
                    for cb in callbacks:
                        cb(child)

                if child['children'] is not None:
                    _walk(child)

        tree = self.root
        _walk(tree)

    def load_data(self, data):
        ''' Load and clean up input data '''

        self.data = utils.clean(data)

    def _make_new_node(self, parent=None, children=None, value=None, path=''):
        ''' Create a new node. If children is NoneType, node is treated as a 
        leaf. 
        '''
        return  dict(parent=parent, children=children, value=value, path=path)

    def _find_ancestor(self, start_node, parents_to_visit):
        ''' Return parent directory corresponding to node parsed from current line. '''

        current_node = start_node

        while (parents_to_visit > 0):
            current_node = current_node['parent']
            parents_to_visit -= 1

        return current_node



# -*- coding: utf-8 -*-

import os
import sys

import utils

def create_file(node):
    ''' Create a regular file if node has NoneType for children.  Otherwise, creates a directory. '''

    if node['path'] == os.path.abspath(os.curdir):
        return

    if os.path.exists(node['path']):
        print('Error, the directory {} already exists'.format(node['path']))
        sys.exit(1) 

    if node['children'] is None:

        try:
            open(node['path'], 'w').close()

        except IOError as E:
            print 'Error: could not create regular file: {}.'.format(node['path'])
            print E
    else:

        try:
            os.mkdir(node['path'])

        except IOError as E:
            print 'Error: could not create directory: {}.'.format(node['path'])
            print E

def make_config_processor(config_path=None):
    ''' Takes relative name of config file in ~/home directory ''' 

    full_path = os.path.abspath( os.path.join(os.path.expanduser('~'), config_path) )

    try:
        with open(full_path) as config_file:
            hooks = dict([  
                            map(str.strip, line.split('=')) 
                            for line in config_file  
                            if line.strip() 
            ])

    except IOError as E:
        print('Could not open config file.')
        print E
        sys.exit(1)

    def process_config_hooks(node):
        ''' on each node, if there's a match in the config settings, that file is created. '''

        filename = node['value'] 
        if filename in hooks:

            try:
                with open( hooks[filename], 'r' ) as src_file, open( node['path'], 'w' ) as dst_file:
                    dst_file.write( src_file.read() ) 

            except IOError as E:
                print('Could not write config hook to new file.')
                print E
                sys.exit(1)

    return process_config_hooks



# -*- coding: utf-8 -*-

'''  
    Utility functions for superdir.py.  
'''

import datetime

import os
import sys

def is_empty(line):
    ''' Return True if line is equal to empty string after being stripped of 
    whitespace. Otherwise, return False.
    '''
    return line.strip() is ''

def is_comment(line):
    return line.strip().startswith('#')

def is_dir(line):
    return line.rstrip().endswith('/')

def is_multiple_of_indent(this_indent, global_indent):
    ''' Returns True if the current indent reaches 0 when the global 
    indent size is repeatedly subtracted from it. Otherwise returns False. 
    '''

    while this_indent > 0:
        this_indent -= global_indent

    return this_indent == 0

def clean(lines):
    ''' Discard all comments and lines with nothing or whitespace. ''' 

    return [ 
            line.rstrip() 
            for line in lines 
            if not ( is_empty(line) or is_comment(line) ) ]

def parse_indent(line):
    ''' Return the leading number of spaces in a line of text. '''

    return len(line) - len(line.lstrip())

def get_indent_count(line, indent_size):
    ''' Return the number of indentation units after dividing a line of 
    text's leading space count by some indent_size. 
    '''
    raw_indent = len(line) - len(line.lstrip())

    rv = None 

    try:
        rv = raw_indent / indent_size

    except ZeroDivisionError:
        rv = 0

    return rv
            
def get_dirname(line):
    ''' Remove all trailing forward slashes from a line after it's been 
    stripped. 
    '''

    return line.strip().rstrip('/')

def get_filename(line):
    ''' Return line with whitespace stripped. '''

    return line.strip()

def get_paths(output_dir):
    ''' Turn /path/to/output_dir into /path/to, output_dir. '''

    abs_cur_dir = os.path.abspath(os.curdir)
    full_base_path = None

    # absolute path example: /data/apps/new_app
    if output_dir.startswith('/'):
        abs_base_path = os.path.join('/', *output_dir.split('/')[:-1])
        output_dir = output_dir.split('/')[-1]

    # relative path with multiple dirs example: apps/new_app 
    else:
        if '/' in output_dir:
            abs_base_path = os.path.join( 
                abs_cur_dir, 
                '/'.join(output_dir.split('/')[:-1]) 
            )
            output_dir = output_dir.split('/')[-1]
        else:
            abs_base_path = abs_cur_dir 

    return abs_base_path, output_dir

def build_output_dirname(dir_suffix='SUPERDIR_OUTPUT', datestring=None):
    ''' Appends a current date in YYYY-MM-DD-HH-MM format to a directory 
    suffix. 
    '''

    if datestring is None:
        import datetime

        dt_now = datetime.datetime.now()
        year, month, day, hour, minute = dt_now.year,\
                                         dt_now.month,\
                                         dt_now.day,\
                                         dt_now.hour,\
                                         dt_now.minute

        datestring = '{}-{}-{}-{}-{}'.format(year, month, day, hour, minute)

    output_dirname = '{}-{}'.format(dir_suffix, datestring)

    return output_dirname

def usage(out=sys.stdout):
    ''' Print usage info. '''

    out.write('Usage: superdir SCHEMA_FILE [TARGET]\n')

def show_err_msg(out=sys.stdout, line_number=None, schema_lines=None):
    ''' Print an error message including the line number and the line. ''' 

    schema_lines[line_number] = schema_lines[line_number].rstrip() +\
                                '    <<< error'
    highlighted_schema = '\n'.join(schema_lines)
    out.write('Parse Error: inconsistent indentation in your schema file '\
              'on line {}\n{}\n'.format(line_number, highlighted_schema))
